@@ -257,15 +257,13 @@ int Grow_addbitmap(char *devname, int fd, char *file, int chunk, int delay, int
 	}
 	bitmapsize = array.size;
 	bitmapsize <<= 1;
-#ifdef BLKGETSIZE64
-	if (ioctl(fd, BLKGETSIZE64, &array_size) == 0 &&
+	if (get_dev_size(fd, NULL, &array_size) &&
 	    array_size > (0x7fffffffULL<<9)) {
 		/* Array is big enough that we cannot trust array.size
 		 * try other approaches
 		 */
 		bitmapsize = get_component_size(fd);
 	}
-#endif
 	if (bitmapsize == 0) {
 		fprintf(stderr, Name ": Cannot reliably determine size of array to create bitmap - sorry.\n");
 		return 1;
@@ -305,9 +303,11 @@ int Grow_addbitmap(char *devname, int fd, char *file, int chunk, int delay, int
 				if (fd2 < 0)
 					continue;
 				if (st->ss->load_super(st, fd2, &super, NULL)==0) {
-					if (st->ss->add_internal_bitmap(st, super,
-									chunk, delay, write_behind,
-									bitmapsize, 0, major))
+					if (st->ss->add_internal_bitmap(
+						    st, super,
+						    &chunk, delay, write_behind,
+						    bitmapsize, 0, major)
+						)
 						st->ss->write_bitmap(st, fd2, super);
 					else {
 						fprintf(stderr, Name ": failed to create internal bitmap - chunksize problem.\n");
@@ -521,7 +521,7 @@ int Grow_reshape(char *devname, int fd, int quiet, char *backup_file,
 	case 6:
 		st = super_by_version(array.major_version,
 				      array.minor_version);
-		/* size can be changed independantly.
+		/* size can be changed independently.
 		 * layout/chunksize/raid_disks/level can be changed
 		 * though the kernel may not support it all.
 		 * If 'suspend_lo' is not present in devfs, then
@@ -618,7 +618,8 @@ int Grow_reshape(char *devname, int fd, int quiet, char *backup_file,
 		printf("mdadm: Need to backup %lluK of critical section..\n", last_block/2);
 
 		sra = sysfs_read(fd, 0,
-				 GET_COMPONENT|GET_DEVS|GET_OFFSET|GET_STATE);
+				 GET_COMPONENT|GET_DEVS|GET_OFFSET|GET_STATE|
+				 GET_CACHE);
 		if (!sra) {
 			fprintf(stderr, Name ": %s: Cannot get array details from sysfs\n",
 				devname);
@@ -657,7 +658,7 @@ int Grow_reshape(char *devname, int fd, int quiet, char *backup_file,
 				offsets[sd->role] = sd->offset;
 				if (fdlist[sd->role] < 0) {
 					fprintf(stderr, Name ": %s: cannot open component %s\n",
-						devname, dn);
+						devname, dn?dn:"-unknown-");
 					goto abort;
 				}
 			} else {
@@ -667,7 +668,7 @@ int Grow_reshape(char *devname, int fd, int quiet, char *backup_file,
 				offsets[d] = sd->offset;
 				if (fdlist[d]<0) {
 					fprintf(stderr, Name ": %s: cannot open component %s\n",
-						devname, dn);
+						devname, dn?dn:"-unknown");
 					goto abort;
 				}
 				d++;
@@ -740,10 +741,25 @@ int Grow_reshape(char *devname, int fd, int quiet, char *backup_file,
 		array.chunk_size = nchunk;
 		array.layout = nlayout;
 		if (ioctl(fd, SET_ARRAY_INFO, &array) != 0) {
+			if (errno == ENOSPC) {
+				/* stripe cache is not big enough.
+				 * It needs to be 4 times chunksize_size,
+				 * and we assume pagesize is 4K
+				 */
+				if (sra->cache_size < 4 * (nchunk/4096)) {
+					sysfs_set_num(sra, NULL,
+						      "stripe_cache_size",
+						      4 * (nchunk/4096) +1);
+					if (ioctl(fd, SET_ARRAY_INFO,
+						  &array) == 0)
+						goto ok;
+				}
+			}
 			fprintf(stderr, Name ": Cannot set device size/shape for %s: %s\n",
 				devname, strerror(errno));
 			goto abort;
 		}
+		ok: ;
 
 		/* suspend the relevant region */
 		sysfs_set_num(sra, NULL, "suspend_hi", 0); /* just in case */
@@ -801,7 +817,10 @@ int Grow_reshape(char *devname, int fd, int quiet, char *backup_file,
 		memset(&bsb, 0, sizeof(bsb));
 		for (i=odisks; i<d ; i++) {
 			lseek64(fdlist[i], (offsets[i]+last_block)<<9, 0);
-			write(fdlist[i], &bsb, sizeof(bsb));
+			if (write(fdlist[i], &bsb, sizeof(bsb)) < 0) {
+				fprintf(stderr, Name ": %s: failed to invalidate metadata for raid disk %d\n",
+					devname, i);
+			}
 		}
 
 		/* unsuspend. */
